const std = @import("std");
const xml = @import("./xml.zig");

// Music Player:
// - Launch VLC and control it via HTTP interface
// - Play music randomly from a VLC playlist
// - Allow media control (play pause, next, prev)
// - Expose OBS Browser Source to display current track information.
// + Add and remove songs from playlist
// + Allow voting on songs (web interface? chat bot?)
//
// Chat bot:
// - Support Twitch chat
// + Support YouTube
//
// A - means immediately, a + means eventually.

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();

    // var vlc = std.process.Child.init(&[_][]const u8{
    //     "vlc",
    //     "--intf",
    //     "http",
    //     "--http-host",
    //     "localhost",
    //     "--http-password",
    //     "1234",
    // }, allocator);
    // try vlc.spawn();

    var tz_file = try std.fs.openFileAbsolute("/etc/localtime", .{});
    defer tz_file.close();
    var tz = try std.tz.Tz.parse(allocator, tz_file.reader());
    defer tz.deinit();

    while (true) : (std.time.sleep(1_000_000_000)) { // sleep 500ms
        try updateTime(tz);
        try updateStatus(allocator);
        scroll += 1;
    }

    // try vlc.kill();
}

const base64_encoder = std.base64.standard.Encoder;

const topic = "Something Fun For Everyone With Streamboy!!!!!!!! git.jeevio.xyz/jeeves/streamboy";

var scroll: usize = 0;

fn updateStatus(allocator: std.mem.Allocator) !void {
    var stream_info_file = try std.fs.createFileAbsolute("/tmp/streaminfo", .{ .truncate = true });
    defer stream_info_file.close();

    var song_info = try getSongInfo(allocator);
    defer song_info.deinit(allocator);

    const string = try std.fmt.allocPrint(allocator, "{s}    |    ♪ {s} - {s}    |    ", .{
        topic,
        song_info.title orelse "Unknown Title",
        song_info.artist orelse "Unknown Artist",
    });
    defer allocator.free(string);

    if (scroll > string.len) scroll = 0;
    if (scroll == 0) try stream_info_file.writeAll(string) else {
        for (string[scroll..]) |char| try stream_info_file.writer().writeByte(char);
        for (string[0..scroll]) |char| try stream_info_file.writer().writeByte(char);
    }
}

const SongInfo = struct {
    title: ?[]const u8,
    album: ?[]const u8,
    artist: ?[]const u8,

    pub fn deinit(self: *SongInfo, allocator: std.mem.Allocator) void {
        if (self.title) |b| allocator.free(b);
        if (self.album) |b| allocator.free(b);
        if (self.artist) |b| allocator.free(b);
    }
};

fn getSongInfo(allocator: std.mem.Allocator) !SongInfo {
    var http = std.http.Client{ .allocator = allocator };
    defer http.deinit();

    const userpass = try std.fmt.allocPrint(allocator, ":{s}", .{"1234"});
    defer allocator.free(userpass);

    const base64_userpass = try allocator.alloc(u8, base64_encoder.calcSize(userpass.len));
    defer allocator.free(base64_userpass);

    const final_userpass = try std.fmt.allocPrint(allocator, "Basic {s}", .{base64_encoder.encode(base64_userpass, userpass)});
    defer allocator.free(final_userpass);

    var response = std.ArrayList(u8).init(allocator);
    defer response.deinit();

    const result = try http.fetch(.{
        .location = .{ .url = "http://localhost:8080/requests/status.xml" },
        .headers = .{ .authorization = .{ .override = final_userpass } },
        .response_storage = .{ .dynamic = &response },
    });

    std.debug.print("{any}\n{s}\n", .{ result, response.items });

    const document = try xml.parse(allocator, response.items);
    defer document.deinit();

    var title: ?[]const u8 = null;
    var album: ?[]const u8 = null;
    var artist: ?[]const u8 = null;

    if (document.root.findChildByTag("information")) |information| {
        var categories_it = information.findChildrenByTag("category");
        while (categories_it.next()) |category| {
            if (std.mem.eql(u8, category.getAttribute("name").?, "meta")) {
                var info_it = category.findChildrenByTag("info");
                while (info_it.next()) |info| {
                    const info_name = info.getAttribute("name").?;
                    if (std.mem.eql(u8, info_name, "title"))
                        title = try processHtmlString(allocator, info.children[0].char_data)
                    else if (std.mem.eql(u8, info_name, "album"))
                        album = try processHtmlString(allocator, info.children[0].char_data)
                    else if (std.mem.eql(u8, info_name, "artist"))
                        artist = try processHtmlString(allocator, info.children[0].char_data);
                }
            }
        }
    }

    return .{
        .title = title,
        .album = album,
        .artist = artist,
    };
}

fn processHtmlString(allocator: std.mem.Allocator, string: []const u8) ![]const u8 {
    var new: []u8 = try allocator.dupe(u8, string);
    errdefer allocator.free(new);
    while (true) {
        if (std.mem.indexOf(u8, new, "&#")) |amp| {
            if (std.mem.indexOfScalarPos(u8, new, amp, ';')) |semi| {
                const int = try std.fmt.parseInt(u8, new[amp + 2 .. semi], 10);
                const nnew = try allocator.alloc(u8, std.mem.replacementSize(u8, new, new[amp .. semi + 1], &[1]u8{int}));
                _ = std.mem.replace(u8, new, new[amp .. semi + 1], &[1]u8{int}, nnew);
                allocator.free(new);
                new = nnew;
            }
        } else break;
    }
    std.debug.print("{s}\n", .{new});
    return new;
}

fn updateTime(tz: std.Tz) !void {
    const original_timestamp = std.time.timestamp();

    var timetype: *std.tz.Timetype = undefined;
    for (tz.transitions, 0..) |trans, i|
        if (trans.ts >= original_timestamp) {
            timetype = tz.transitions[i - 1].timetype;
            break;
        };

    const timestamp = std.time.timestamp() + timetype.offset + std.time.s_per_day;

    const epoch_seconds = std.time.epoch.EpochSeconds{ .secs = @intCast(timestamp) };
    const day_seconds = epoch_seconds.getDaySeconds();
    const epoch_day = epoch_seconds.getEpochDay();
    const year_day = epoch_day.calculateYearDay();
    const month_day = year_day.calculateMonthDay();

    const ampm = getAmPm(day_seconds);

    var file = try std.fs.createFileAbsolute("/tmp/time", .{ .truncate = true });
    defer file.close();

    try file.writer().print("{s}, {s} {d} {d}, {d:0>2}:{d:0>2}:{d:0>2} {s}", .{
        getDayOfWeekName(getDayOfWeek(epoch_day)),
        getMonthName(month_day.month),
        month_day.day_index,
        year_day.year,
        ampm.hour,
        day_seconds.getMinutesIntoHour(),
        day_seconds.getSecondsIntoMinute(),
        if (ampm.is_pm) "PM" else "AM",
    });
}

fn getMonthName(month: std.time.epoch.Month) []const u8 {
    return switch (month) {
        .jan => "January",
        .feb => "February",
        .mar => "March",
        .apr => "April",
        .may => "May",
        .jun => "June",
        .jul => "July",
        .aug => "August",
        .sep => "September",
        .oct => "October",
        .nov => "November",
        .dec => "December",
    };
}

// Jan 1 1970 was a Thursday, so we make that 1
const DayOfWeek = enum(u3) {
    wed = 0,
    thu,
    fri,
    sat,
    sun,
    mon,
    tue,
};

fn getDayOfWeek(epoch_day: std.time.epoch.EpochDay) DayOfWeek {
    return @enumFromInt(@mod(epoch_day.day, 7));
}

fn getDayOfWeekName(day_of_week: DayOfWeek) []const u8 {
    return switch (day_of_week) {
        .sun => "Sunday",
        .mon => "Monday",
        .tue => "Tuesday",
        .wed => "Wednesday",
        .thu => "Thursday",
        .fri => "Friday",
        .sat => "Saturday",
    };
}

const AmPm = struct {
    hour: u4,
    is_pm: bool,
};

fn getAmPm(day_seconds: std.time.epoch.DaySeconds) AmPm {
    const hour = day_seconds.getHoursIntoDay();
    return .{
        .hour = if (hour < 13) @intCast(hour) else @intCast(hour - 12),
        .is_pm = hour != 0 and hour > 11,
    };
}
